from typing import Any
from tomodachi.protocol.json_base import JsonBase
try:
    from tomodachi.protocol.protobuf_base import ProtobufBase  # type: ignore
except Exception:
    class ProtobufBase(object):  # type: ignore
        @classmethod
        def validate(cls, **kwargs: Any) -> None:
            raise Exception('google.protobuf package not installed')

__all__ = ['JsonBase', 'ProtobufBase']



import re


class RegexMissmatchException(Exception):
    def __init__(self, value: str, pattern: str) -> None:
        message = 'RegexMissmatchException: value "{}" does not match pattern "{}"'.format(value, pattern)
        super().__init__(message)


def validate_field_regex(value: str, pattern_str: str) -> None:
    pattern = re.compile(pattern_str)
    if not pattern.match(value):
        raise RegexMissmatchException(value=value, pattern=pattern_str)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protobuf/sns_sqs_message.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='protobuf/sns_sqs_message.proto',
  package='proto',
  syntax='proto3',
  serialized_pb=_b('\n\x1eprotobuf/sns_sqs_message.proto\x12\x05proto\"%\n\x07Service\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04uuid\x18\x02 \x01(\t\"\x99\x01\n\x08Metadata\x12\x14\n\x0cmessage_uuid\x18\x01 \x01(\t\x12\x18\n\x10protocol_version\x18\x02 \x01(\t\x12$\n\x1c\x63ompatible_protocol_versions\x18\x03 \x03(\t\x12\x11\n\ttimestamp\x18\x04 \x01(\x01\x12\r\n\x05topic\x18\x05 \x01(\t\x12\x15\n\rdata_encoding\x18\x06 \x01(\t\"a\n\rSNSSQSMessage\x12\x1f\n\x07service\x18\x01 \x01(\x0b\x32\x0e.proto.Service\x12!\n\x08metadata\x18\x02 \x01(\x0b\x32\x0f.proto.Metadata\x12\x0c\n\x04\x64\x61ta\x18\x03 \x01(\tb\x06proto3')
)




_SERVICE = _descriptor.Descriptor(
  name='Service',
  full_name='proto.Service',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='proto.Service.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='uuid', full_name='proto.Service.uuid', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=41,
  serialized_end=78,
)


_METADATA = _descriptor.Descriptor(
  name='Metadata',
  full_name='proto.Metadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='message_uuid', full_name='proto.Metadata.message_uuid', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='protocol_version', full_name='proto.Metadata.protocol_version', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='compatible_protocol_versions', full_name='proto.Metadata.compatible_protocol_versions', index=2,
      number=3, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='proto.Metadata.timestamp', index=3,
      number=4, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='topic', full_name='proto.Metadata.topic', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_encoding', full_name='proto.Metadata.data_encoding', index=5,
      number=6, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=81,
  serialized_end=234,
)


_SNSSQSMESSAGE = _descriptor.Descriptor(
  name='SNSSQSMessage',
  full_name='proto.SNSSQSMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='proto.SNSSQSMessage.service', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='metadata', full_name='proto.SNSSQSMessage.metadata', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='proto.SNSSQSMessage.data', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=236,
  serialized_end=333,
)

_SNSSQSMESSAGE.fields_by_name['service'].message_type = _SERVICE
_SNSSQSMESSAGE.fields_by_name['metadata'].message_type = _METADATA
DESCRIPTOR.message_types_by_name['Service'] = _SERVICE
DESCRIPTOR.message_types_by_name['Metadata'] = _METADATA
DESCRIPTOR.message_types_by_name['SNSSQSMessage'] = _SNSSQSMESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Service = _reflection.GeneratedProtocolMessageType('Service', (_message.Message,), dict(
  DESCRIPTOR = _SERVICE,
  __module__ = 'protobuf.sns_sqs_message_pb2'
  # @@protoc_insertion_point(class_scope:proto.Service)
  ))
_sym_db.RegisterMessage(Service)

Metadata = _reflection.GeneratedProtocolMessageType('Metadata', (_message.Message,), dict(
  DESCRIPTOR = _METADATA,
  __module__ = 'protobuf.sns_sqs_message_pb2'
  # @@protoc_insertion_point(class_scope:proto.Metadata)
  ))
_sym_db.RegisterMessage(Metadata)

SNSSQSMessage = _reflection.GeneratedProtocolMessageType('SNSSQSMessage', (_message.Message,), dict(
  DESCRIPTOR = _SNSSQSMESSAGE,
  __module__ = 'protobuf.sns_sqs_message_pb2'
  # @@protoc_insertion_point(class_scope:proto.SNSSQSMessage)
  ))
_sym_db.RegisterMessage(SNSSQSMessage)


# @@protoc_insertion_point(module_scope)



import datetime
import asyncio
import time
import pytz
import tzlocal
import traceback
import inspect
import logging
from typing import Any, Dict, List, Union, Optional, Callable, Tuple, Awaitable  # noqa
from tomodachi.invoker import Invoker
from tomodachi.helpers.crontab import get_next_datetime


class Scheduler(Invoker):
    close_waiter = None

    async def schedule_handler(cls: Any, obj: Any, context: Dict, func: Any, interval: Optional[Union[str, int]]=None, timestamp: Optional[str]=None, timezone: Optional[str]=None, immediately: Optional[bool]=False) -> Any:
        async def handler() -> None:
            values = inspect.getfullargspec(func)
            kwargs = {k: values.defaults[i] for i, k in enumerate(values.args[len(values.args) - len(values.defaults):])} if values.defaults else {}
            try:
                routine = func(*(obj,), **kwargs)
                if isinstance(routine, Awaitable):
                    await routine
            except Exception as e:
                if not context.get('log_level') or context.get('log_level') in ['DEBUG']:
                    traceback.print_exception(e.__class__, e, e.__traceback__)

        context['_schedule_scheduled_functions'] = context.get('_schedule_scheduled_functions', [])
        context['_schedule_scheduled_functions'].append((interval, timestamp, timezone, immediately, func, handler))

        start_func = cls.start_scheduler(cls, obj, context)
        return (await start_func) if start_func else None

    @classmethod
    def schedule_handler_with_interval(cls, interval: Union[str, int]) -> Callable:
        def _func(cls: Any, obj: Any, context: Dict, func: Any) -> Any:
            return cls.schedule_handler(cls, obj, context, func, interval=interval)

        return _func

    @classmethod
    def next_call_at(cls, current_time: float, interval: Optional[Union[str, int]]=None, timestamp: Optional[str]=None, timezone: Optional[str]=None) -> int:
        if not timezone:
            tz = tzlocal.get_localzone()
        else:
            try:
                tz = pytz.timezone(timezone or '')
            except Exception as e:
                raise Exception('Unknown timezone: {}'.format(timezone)) from e
        local_tz = tzlocal.get_localzone()

        if interval is None and timestamp is not None:
            if isinstance(timestamp, str):
                try:
                    datetime_object = datetime.datetime.strptime(timestamp, '%Y-%m-%d %H:%M:%S')
                    interval = '{} {} {} {} * {}'.format(datetime_object.minute, datetime_object.hour, datetime_object.day, datetime_object.month, datetime_object.year)
                    second_modifier = 1
                    if tz.localize(datetime_object) > local_tz.localize(datetime.datetime.fromtimestamp(current_time)):
                        second_modifier = -60
                    next_at = get_next_datetime(interval, local_tz.localize(datetime.datetime.fromtimestamp(current_time + second_modifier)).astimezone(tz))
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp() + datetime_object.second)
                except ValueError:
                    pass

                try:
                    datetime_object = datetime.datetime.strptime(timestamp, '%Y-%m-%d %H:%M')
                    interval = '{} {} {} {} * {}'.format(datetime_object.minute, datetime_object.hour, datetime_object.day, datetime_object.month, datetime_object.year)
                    next_at = get_next_datetime(interval, local_tz.localize(datetime.datetime.fromtimestamp(current_time + 1)).astimezone(tz))
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp())
                except ValueError:
                    pass

                try:
                    datetime_object = datetime.datetime.strptime(timestamp, '%H:%M:%S')
                    datetime_object = datetime.datetime(datetime.datetime.fromtimestamp(current_time).year, datetime.datetime.fromtimestamp(current_time).month, datetime.datetime.fromtimestamp(current_time).day, datetime_object.hour, datetime_object.minute, datetime_object.second)
                    interval = '{} {} * * *'.format(datetime_object.minute, datetime_object.hour)
                    second_modifier = 1
                    if tz.localize(datetime_object) > local_tz.localize(datetime.datetime.fromtimestamp(current_time)):
                        second_modifier = -60
                    next_at = get_next_datetime(interval, local_tz.localize(datetime.datetime.fromtimestamp(current_time + second_modifier)).astimezone(tz))
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp() + datetime_object.second)
                except ValueError:
                    pass

                try:
                    datetime_object = datetime.datetime.strptime(timestamp, '%H:%M')
                    interval = '{} {} * * *'.format(datetime_object.minute, datetime_object.hour)
                    next_at = get_next_datetime(interval, local_tz.localize(datetime.datetime.fromtimestamp(current_time + 1)).astimezone(tz))
                    if not next_at:
                        return int(current_time + 60 * 60 * 24 * 365 * 100)
                    return int(next_at.timestamp())
                except ValueError:
                    pass

                raise Exception('Invalid timestamp')

        if interval is not None:
            if isinstance(interval, int):
                return int(current_time + interval)

            interval_aliases = {
                ('every second', '1s', '1 s', '1second', '1 second', 'second', 'secondly', 'once per second'): 1,
                ('every minute', '1m', '1 m', '1minute', '1 minute', 'minute', 'minutely', 'once per minute'): '@minutely',
                ('every hour', '1h', '1 h', '1hour', '1 hour', 'hour', 'hourly', 'once per hour'): '@hourly',
                ('every day', '1d', '1 d', '1day', '1 day', 'day', 'daily', 'once per day', 'nightly'): '@daily',
                ('every month', '1month', '1 month', 'month', 'monthly', 'once per month'): '@monthly',
                ('every year', '1y', '1 y', '1year', '1 year', 'year', 'yearly', 'once per year', 'annually'): '@yearly',
                ('monday', 'mondays', 'mon', 'every monday', 'once per monday', 'weekly', 'once per week', 'week', 'every week'): '0 0 * * 1',
                ('tuesday', 'tuesdays', 'tue', 'every tuesday', 'once per tuesday'): '0 0 * * 2',
                ('wednesday', 'wednesdays', 'wed', 'every wednesday', 'once per wednesday'): '0 0 * * 3',
                ('thursday', 'thursdays', 'thu', 'every thursday', 'once per thursday'): '0 0 * * 4',
                ('friday', 'fridays', 'fri', 'every friday', 'once per friday'): '0 0 * * 5',
                ('saturday', 'saturdays', 'sat', 'every saturday', 'once per saturday'): '0 0 * * 6',
                ('sunday', 'sundays', 'sun', 'every sunday', 'once per sunday'): '0 0 * * 0',
                ('weekday', 'weekdays', 'every weekday'): '0 0 * * 1-5',
                ('weekend', 'weekends', 'every weekend'): '0 0 * * 0,6'
            }  # type: Dict[Tuple[str, ...], Union[str, int]]
            interval = interval.lower()

            if interval.endswith('s') or interval.endswith('seconds'):
                try:
                    interval = int(interval.replace('seconds', '').replace('s', '').replace(' ', ''))
                except ValueError:
                    pass

            try:
                interval_value = [v for k, v in interval_aliases.items() if interval in k][0]  # type: Union[str, int]
            except IndexError:
                interval_value = interval
            if isinstance(interval_value, int):
                return int(current_time + interval_value)

            try:
                next_at = get_next_datetime(interval_value, local_tz.localize(datetime.datetime.fromtimestamp(current_time + 1)).astimezone(tz))
                if not next_at:
                    return int(current_time + 60 * 60 * 24 * 365 * 100)
                return int(next_at.timestamp())
            except Exception:
                raise Exception('Invalid interval')

        return int(current_time + 60 * 60 * 24 * 365 * 100)

    def get_timezone(cls: Any, timezone: Optional[str]=None) -> Optional[str]:
        if timezone:
            tz_aliases = {
                ('+00:00', '-00:00', '00:00', '0000', 'GMT +0000', 'GMT +00:00', 'GMT -00', 'GMT +00', 'GMT -0', 'GMT +0'): 'GMT0',
                ('+01:00', '+0100', 'GMT +0100', 'GMT +01:00', 'GMT +01', 'GMT +1'): 'Etc/GMT-1',
                ('+02:00', '+0200', 'GMT +0200', 'GMT +02:00', 'GMT +02', 'GMT +2'): 'Etc/GMT-2',
                ('+03:00', '+0300', 'GMT +0300', 'GMT +03:00', 'GMT +03', 'GMT +3'): 'Etc/GMT-3',
                ('+04:00', '+0400', 'GMT +0400', 'GMT +04:00', 'GMT +04', 'GMT +4'): 'Etc/GMT-4',
                ('+05:00', '+0500', 'GMT +0500', 'GMT +05:00', 'GMT +05', 'GMT +5'): 'Etc/GMT-5',
                ('+06:00', '+0600', 'GMT +0600', 'GMT +06:00', 'GMT +06', 'GMT +6'): 'Etc/GMT-6',
                ('+07:00', '+0700', 'GMT +0700', 'GMT +07:00', 'GMT +07', 'GMT +7'): 'Etc/GMT-7',
                ('+08:00', '+0800', 'GMT +0800', 'GMT +08:00', 'GMT +08', 'GMT +8'): 'Etc/GMT-8',
                ('+09:00', '+0900', 'GMT +0900', 'GMT +09:00', 'GMT +09', 'GMT +9'): 'Etc/GMT-9',
                ('+10:00', '+1000', 'GMT +1000', 'GMT +10:00', 'GMT +10'): 'Etc/GMT-10',
                ('+11:00', '+1100', 'GMT +1100', 'GMT +11:00', 'GMT +11'): 'Etc/GMT-11',
                ('+12:00', '+1200', 'GMT +1200', 'GMT +12:00', 'GMT +12'): 'Etc/GMT-12',
                ('-01:00', '-0100', 'GMT -0100', 'GMT -01:00', 'GMT -01', 'GMT -1'): 'Etc/GMT+1',
                ('-02:00', '-0200', 'GMT -0200', 'GMT -02:00', 'GMT -02', 'GMT -2'): 'Etc/GMT+2',
                ('-03:00', '-0300', 'GMT -0300', 'GMT -03:00', 'GMT -03', 'GMT -3'): 'Etc/GMT+3',
                ('-04:00', '-0400', 'GMT -0400', 'GMT -04:00', 'GMT -04', 'GMT -4'): 'Etc/GMT+4',
                ('-05:00', '-0500', 'GMT -0500', 'GMT -05:00', 'GMT -05', 'GMT -5'): 'Etc/GMT+5',
                ('-06:00', '-0600', 'GMT -0600', 'GMT -06:00', 'GMT -06', 'GMT -6'): 'Etc/GMT+6',
                ('-07:00', '-0700', 'GMT -0700', 'GMT -07:00', 'GMT -07', 'GMT -7'): 'Etc/GMT+7',
                ('-08:00', '-0800', 'GMT -0800', 'GMT -08:00', 'GMT -08', 'GMT -8'): 'Etc/GMT+8',
                ('-09:00', '-0900', 'GMT -0900', 'GMT -09:00', 'GMT -09', 'GMT -9'): 'Etc/GMT+9',
                ('-10:00', '-1000', 'GMT -1000', 'GMT -10:00', 'GMT -10'): 'Etc/GMT+10',
                ('-11:00', '-1100', 'GMT -1100', 'GMT -11:00', 'GMT -11'): 'Etc/GMT+11',
                ('-12:00', '-1200', 'GMT -1200', 'GMT -12:00', 'GMT -12'): 'Etc/GMT+12'
            }  # type: Dict[Tuple[str, ...], str]
            try:
                try:
                    timezone = [v for k, v in tz_aliases.items() if timezone in k or timezone.replace(' ', '') in [x.replace(' ', '') for x in k]][0]
                except IndexError:
                    pass
                pytz.timezone(timezone or '')
            except Exception as e:
                raise Exception('Unknown timezone: {}'.format(timezone)) from e

        return timezone

    async def start_schedule_loop(cls: Any, obj: Any, context: Dict, handler: Callable, func: Callable, interval: Optional[Union[str, int]]=None, timestamp: Optional[str]=None, timezone: Optional[str]=None, immediately: Optional[bool]=False) -> None:
        timezone = cls.get_timezone(timezone)

        if not cls.close_waiter:
            cls.close_waiter = asyncio.Future()
        stop_waiter = asyncio.Future()  # type: asyncio.Future
        start_waiter = asyncio.Future()  # type: asyncio.Future

        async def schedule_loop() -> None:
            await start_waiter
            next_call_at = None
            prev_call_at = None
            tasks = []  # type: List
            current_time = time.time()
            too_many_tasks = False
            run_immediately = immediately
            while not cls.close_waiter.done():
                try:
                    if not run_immediately:
                        last_time = current_time
                        actual_time = time.time()
                        current_time = last_time + 1 if int(last_time + 1) < int(actual_time) else actual_time
                        if next_call_at is None:
                            next_call_at = cls.next_call_at(current_time, interval, timestamp, timezone)
                            if prev_call_at and prev_call_at == next_call_at:
                                next_call_at = None
                                continue
                        sleep_diff = int(current_time + 1) - actual_time + 0.001
                        if sleep_diff > 0:
                            await asyncio.sleep(sleep_diff)
                        if next_call_at > time.time():
                            continue
                    run_immediately = False
                    if cls.close_waiter.done():
                        continue
                    prev_call_at = next_call_at
                    next_call_at = None
                    tasks = [task for task in tasks if not task.done()]
                    if len(tasks) >= 20:
                        if not too_many_tasks:
                            too_many_tasks = True
                            logging.getLogger('transport.schedule').warning('Too many scheduled tasks (20) for function "{}"'.format(func.__name__))
                        await asyncio.sleep(1)
                        continue
                    too_many_tasks = False
                    tasks.append(asyncio.ensure_future(asyncio.shield(handler())))
                except Exception as e:
                    if not context.get('log_level') or context.get('log_level') in ['DEBUG']:
                        traceback.print_exception(e.__class__, e, e.__traceback__)
                    await asyncio.sleep(1)

            if tasks:
                await asyncio.wait(tasks)
            if not stop_waiter.done():
                stop_waiter.set_result(None)

        loop = asyncio.get_event_loop()  # type: Any

        stop_method = getattr(obj, '_stop_service', None)

        async def stop_service(*args: Any, **kwargs: Any) -> None:
            if not cls.close_waiter.done():
                cls.close_waiter.set_result(None)

                if not start_waiter.done():
                    start_waiter.set_result(None)
                await stop_waiter
                if stop_method:
                    await stop_method(*args, **kwargs)
            else:
                await stop_waiter
                if stop_method:
                    await stop_method(*args, **kwargs)

        setattr(obj, '_stop_service', stop_service)

        started_method = getattr(obj, '_started_service', None)

        async def started_service(*args: Any, **kwargs: Any) -> None:
            if started_method:
                await started_method(*args, **kwargs)
            if not start_waiter.done():
                start_waiter.set_result(None)

        setattr(obj, '_started_service', started_service)

        loop.create_task(schedule_loop())

    async def start_scheduler(cls: Any, obj: Any, context: Dict) -> Optional[Callable]:
        if context.get('_schedule_loop_started'):
            return None
        context['_schedule_loop_started'] = True

        async def _schedule() -> None:
            cls.close_waiter = asyncio.Future()

            for interval, timestamp, timezone, immediately, func, handler in context.get('_schedule_scheduled_functions', []):
                cls.next_call_at(time.time(), interval, timestamp, cls.get_timezone(timezone))  # test provided interval/timestamp on init

            for interval, timestamp, timezone, immediately, func, handler in context.get('_schedule_scheduled_functions', []):
                await cls.start_schedule_loop(cls, obj, context, handler, func, interval, timestamp, timezone, immediately)

        return _schedule


schedule = Scheduler.decorator(Scheduler.schedule_handler)

heartbeat = Scheduler.decorator(Scheduler.schedule_handler_with_interval(1))
every_second = Scheduler.decorator(Scheduler.schedule_handler_with_interval(1))

minutely = Scheduler.decorator(Scheduler.schedule_handler_with_interval('minutely'))
hourly = Scheduler.decorator(Scheduler.schedule_handler_with_interval('hourly'))
daily = Scheduler.decorator(Scheduler.schedule_handler_with_interval('daily'))
monthly = Scheduler.decorator(Scheduler.schedule_handler_with_interval('monthly'))



import logging
import uuid
import time
import base64
from typing import Any, Dict, Tuple, Union

from tomodachi.proto_build.protobuf.sns_sqs_message_pb2 import SNSSQSMessage

PROTOCOL_VERSION = 'protobuf_base-wip'
COMPATIBLE_PROTOCOL_VERSIONS = ['protobuf_base-wip']


class ProtobufBase(object):
    @classmethod
    def validate(cls, **kwargs: Any) -> None:
        if 'proto_class' not in kwargs:
            raise Exception('No proto_class defined')
        if kwargs.get('proto_class', None).__class__.__name__ != 'GeneratedProtocolMessageType':
            raise Exception('proto_class is not a GeneratedProtocolMessageType')

    @classmethod
    async def build_message(cls, service: Any, topic: str, data: Any) -> bytes:
        message = SNSSQSMessage()
        message.service.name = getattr(service, 'name', None)
        message.service.uuid = getattr(service, 'uuid', None)
        message.metadata.message_uuid = '{}.{}'.format(getattr(service, 'uuid', ''), str(uuid.uuid4()))
        message.metadata.protocol_version = PROTOCOL_VERSION
        message.metadata.compatible_protocol_versions.extend(COMPATIBLE_PROTOCOL_VERSIONS)
        message.metadata.timestamp = time.time()
        message.metadata.topic = topic
        message.metadata.data_encoding = 'base64'
        message.data = base64.b64encode(data.SerializeToString())
        return base64.b64encode(message.SerializeToString())

    @classmethod
    async def parse_message(cls, payload: str, proto_class: Any, validator: Any = None) -> Union[Dict, Tuple]:
        message = SNSSQSMessage()
        message.ParseFromString(base64.b64decode(payload))

        message_uuid = message.metadata.message_uuid
        timestamp = message.metadata.timestamp

        if PROTOCOL_VERSION not in message.metadata.compatible_protocol_versions:
            return False, message_uuid, timestamp

        obj = proto_class()
        obj.ParseFromString(base64.b64decode(message.data))

        if validator is not None:
            try:
                if hasattr(validator, '__func__'):
                    # for static functions
                    validator.__func__(obj)
                else:
                    # for non-static functions
                    validator(obj)
            except Exception as e:
                logging.getLogger('protocol.protobuf_base').warning(e.__str__())
                raise e

        return {
            'service': {
                'name': message.service.name,
                'uuid': message.service.uuid
            },
            'metadata': {
                'message_uuid': message.metadata.message_uuid,
                'protocol_version': message.metadata.protocol_version,
                'compatible_protocol_versions': message.metadata.compatible_protocol_versions,
                'timestamp': message.metadata.timestamp,
                'topic': message.metadata.topic,
                'data_encoding': message.metadata.data_encoding
            },
            'data': obj
        }, message_uuid, timestamp
